<?php

/**
 * @file
 * Provides extended_file_field integration for the PIFT module.
 */

/**
 * Implements hook_extended_file_field_metadata_types().
 *
 * Defines additional PIFT metadata properties for the extended file field
 * formatter and widget, including the associated 'Test ID', 'Test Status',
 * 'Test Message' and 'Test Operations' information for a given file.
 */
function pift_extended_file_field_metadata_types() {
  return array(
    'pift_id' => array(
      'title' => t('Test ID'),
      'sort' => 'numeric',
      'formatter' => 'pift_extended_file_field_testid_formatter',
    ),
    'pift_status' => array(
      'title' => t('Test Status'),
      'sort' => 'string',
      'formatter' => 'pift_extended_file_field_status_formatter',
    ),
    'pift_message' => array(
      'title' => t('Test Summary'),
      'sort' => 'string',
      'formatter' => 'pift_extended_file_field_message_formatter',
    ),
    'pift_operations' => array(
      'title' => t('Operations'),
      'sort' => FALSE,
      'formatter' => 'pift_extended_file_field_operations_formatter',
    ),
  );
}

/**
 * Implementation of hook_extended_file_field_items_alter().
 *
 * Adds PIFT related metadata (including test id, test result, and test
 * operations) to the file $items array within fields using the
 * extended_file_field field formatter.
 *
 * @param array $items
 *   The array of file items from the field, keyed by file ID (fid).
 * @param array $context
 *   Associative array of context for the table of files being altered, with
 *   the following keys:
 *   - field: The field definition array.
 *   - instance: The field instance definition array.
 *   - entity: An object representing the entity the file field is attached to.
 *   - entity_type: String with the type of entity the field is attached to.
 *   - langcode: The language associated with $items.
 *   - display: The display settings to use, as found in the 'display' entry of
 *     the instance definition. Notable keys include the name of the formatter
 *     (in 'type') and the array of formatter settings (in 'settings').
 *
 *  @see drupal_alter()
 */
function pift_extended_file_field_items_alter(&$items, &$context) {
  // Add a 'Test ID' property to each item in the $items array.
  $fids = array_keys($items);
  // If there are no files, bail.
  if (empty($fids)) {
    return;
  }
  $node = $context['entity'];
  // If there are no test results on this node, bail.
  if (empty($node->test_results)) {
    return;
  }

  $retest = user_access('pift re-test files');

  foreach ($items as $key => $value) {
    if (isset($node->test_results[$key])) {
      // Add the test id associated with each file to the $items array
      $items[$key]['pift_id'] = $node->test_results[$key]['test_id'];
      $items[$key]['pift_status'] = $node->test_results[$key]['status'];
      $items[$key]['pift_message'] = $node->test_results[$key]['message'];
      $items[$key]['pift_operations'] = array('view' => $node->test_results[$key]['test_id']);
      if ($retest && in_array($node->test_results[$key]['status'], array(PIFT_STATUS_FAIL, PIFT_STATUS_PASS))) {
        $items[$key]['pift_operations']['retest'] = $node->test_results[$key]['test_id'];
      }
    }
    else {
      // If no test result is found for a given file, set all values to '0'.
      $items[$key]['pift_id'] = 0;
      $items[$key]['pift_status'] = 0;
      $items[$key]['pift_message'] = 0;
      $items[$key]['pift_operations'] = array();
    }
  }
}

/**
 * Implementation of hook_extended_file_field_widget_items_alter().
 *
 * Adds PIFT related metadata (including test id, test result, and test
 * operations) to the file $items array within fields using the
 * extended_file_field field widget.
 *
 * @param array $items
 *   The array of file items from the field, keyed by file ID (fid).
 * @param array $context
 *   Associative array of context for the table of files being altered, with
 *   the following keys:
 *   - field: The field definition array.
 *   - instance: The field instance definition array.
 *   - entity: An object representing the entity the file field is attached to.
 *   - entity_type: String with the type of entity the field is attached to.
 *   - langcode: The language associated with $items.
 *   - display: The display settings to use, as found in the 'display' entry of
 *     the instance definition. Notable keys include the name of the formatter
 *     (in 'type') and the array of formatter settings (in 'settings').
 *
 *  @see drupal_alter()
 */
function pift_extended_file_field_widget_items_alter(&$items, &$context) {
  pift_extended_file_field_items_alter($items, $context);
}

/**
 * Generates the 'file test_id' column for a given file.
 *
 * If the file has a valid test entry in the 'pift_id' property, this
 * function renders it.
 */
function pift_extended_file_field_testid_formatter($file, $context) {
  return !empty($file['pift_id']) ? check_plain($file['pift_id']) : '';
}


/**
 * Generates the 'file status' column for a given file.
 *
 * If the file has a valid test entry in the 'pift_status' property, this
 * function renders the human-readable version of the test status.
 */
function pift_extended_file_field_status_formatter($file, $context) {
  if (!empty($file['pift_id'])) {
    module_load_include('inc', 'pift', 'pift.pages');
    list($cell, $class) = pift_attachment_process($file['pift_status'], (bool) $file['pift_message']);
    $data = $cell;
  }
  return (!empty($data)) ? $data : '';
}

/**
 * Generates the 'file message' column for a given file.
 *
 * If the file has a valid test entry in the 'pift_status' property, this
 * function renders the human-readable version of the test status.
 */
function pift_extended_file_field_message_formatter($file, $context) {
  if (!empty($file['pift_id'])) {
    if ($file['pift_status'] == PIFT_STATUS_QUEUE) {
      $message = t('Queued for testing.');
      if (!empty($file['pift_message'])) {
        $message .= '<div class="pift-previous">' . t('Previous result: @message', array('@message' => $file['pift_message'])) . '</div>';
      }
    }
    elseif ($file['pift_status'] == PIFT_STATUS_SENT) {
      $message = t('Test request sent.');
      if (!empty($file['pift_message'])) {
        $message .= '<div class="pift-previous">' . t('Previous result: @message', array('@message' => $file['pift_message'])) . '</div>';
      }
    }
    else {
      $message = $file['pift_message'] ? check_plain($file['pift_message']) : '<em>' . t('None') . '</em>';
    }
  }
  return (!empty($message)) ? $message : '';
}

/**
 * Generates the 'file operations' column for a given file.
 *
 * If the file has a valid test entry in the pift_data table, this function
 * renders the 'view test' and 're-test' links used to manage testing for the
 * file.
 */
function pift_extended_file_field_operations_formatter($file, $context) {
  if (!empty($file['pift_id'])) {
    foreach ($file['pift_operations'] as $op => $test_id) {
      if ($op == 'view') {
        $links[] = pift_generate_view_link($test_id);
      }
      elseif ($op == 'retest') {
        $links[] = pift_generate_retest_link($test_id);
      }
    }
  }
  return (!empty($links)) ? implode(' | ', $links) : '';
}

/**
 * Generates the 'view details' link for a given test_id.
 */
function pift_generate_view_link($test_id) {
  return l(t('View'), variable_get('pift_server', 'http://example.com') . 'pifr/test/' . $test_id);
}

/**
 * Generates the 'Re-test' link for a given test_id.
 */
function pift_generate_retest_link($test_id) {
  return l(t('Retest'), 'pift/retest/' . $test_id);
}

/**
 * Generates the 'Delete' link for a given test_id.
 */
function pift_generate_delete_link($test_id) {
  return l(t('Delete'), 'pift/delete/' . $test_id);
}

/**
 * Implements hook_extended_file_field_output_alter().
 *
 * Modifies the extended_file_field table structure to add PIFT test results.
 */
function pift_extended_file_field_output_alter(&$elements, &$context) {
  // If we have not been passed any rows to add results to, then bail.
  if (empty($elements[0]['#rows'])) {
    return;
  }
  $items = $context['items'];
  $node = $context['entity'];
  $retest = user_access('pift re-test files');
  $classes = array('', 'pift-queue', 'pift-retest', 'pift-fail', 'pift-pass');
  $output_rows = array();
  $flip = array('odd' => 'even', 'even' => 'odd');
  $striping = 'odd';
  // Cycle through each row, adding PIFT details as needed.
  foreach ($elements[0]['#rows'] as $fid => $row) {

    // Group files with the same comment id, applying similar table striping.
    if (!empty($last_cid) && $row['data']['cid']['data'] != $last_cid) {
      $striping = $flip[$striping];
    }

    // Add the original row
    $output_rows[$fid] = $row;

    // Set no_striping to allow us to override table striping
    $output_rows[$fid]['no_striping'] = TRUE;
    $output_rows[$fid]['class'][] = $striping;

    // Add the PIFT result details
    if (!empty($node->test_results[$fid])) {

      // Set a class and rowspan on the comment id column
      if (!empty($row['data']['cid'])) {
        $output_rows[$fid]['data']['cid']['rowspan'] = 2;
      }

      $column_count = isset($row['data']) ? count($row['data']) : count($row);
      $message = pift_extended_file_field_message_formatter($items[$fid], array());
      $operations = pift_extended_file_field_operations_formatter($items[$fid], array());
      if (!empty($operations)) {
        $message .= '<div class="pift-operations">[ ' .  $operations . ' ]</div>';
      }
      $output_rows["$fid-results"] = array(
        'data' => array(
          array(
            'data' => $message,
            'class' => array($classes[$items[$fid]['pift_status']]),
            'colspan' => $column_count - 1,
          ),
        ),
        'no_striping' => TRUE,
        'class' => array_unique(array_merge(array('extended-file-field-table-result-row', 'pift-test-info', $striping), $row['class'])),
      );
    }
    // Set the $last_cid parameter for table striping
    $last_cid = !empty($row['data']['cid']['data']) ? $row['data']['cid']['data'] : NULL;
  }

  // Override the table rows with our updated element.
  $elements[0]['#rows'] = $output_rows;
}
