<?php

/**
 * @file
 * Provides general pages and UI functions for the PIFT module.
 */

/**
 * Menu page callback: Generates the 'Automated Testing' tab on project nodes.
 *
 * @param $node
 *   The fully loaded project node object.
 *
 * @return $content
 *   A render array representing the output for this page.
 */
function pift_pages_project_testing_tab($node) {

  // Generate the introductory banner for the page.
  $content['banner'] = array(
    '#markup' => '<p>' . t("This page provides information regarding automated testing status for this project's releases.") . '</p>',
  );

  // Generate the project_issue_testing_settings form.
  $content['settings'] = drupal_get_form('pift_pages_project_testing_settings', $node);

  // Generate the project_issue_testing_results table.
  $content['results'] = pift_pages_project_testing_results($node);

  return $content;
}

/**
 * Generate a table of test results for the current project node.
 *
 * @param $node
 *   The fully loaded project node object.
 *
 * @return $content
 *   A render array defining the test results table for this project.
 */
function pift_pages_project_testing_results($node) {
  // Load an array of releases available in this project's repository.
  // We know we will have at least one branch, due to the 'has releases' check
  // in the access function for this callback.
  $releases = project_release_query_releases($node->nid, NULL, NULL, TRUE, TRUE);

  $compatible = pift_core_api_versions();
  $api_vocabulary = taxonomy_vocabulary_load(variable_get('project_release_api_vocabulary', ''));
  $api_field = 'taxonomy_' . $api_vocabulary->machine_name;

  $branches = array();
  foreach ($releases as $rid => $release) {
    if (!isset($release->versioncontrol_release) || !isset($release->{$api_field})) {
      continue;
    }
    if (!in_array($release->{$api_field}[LANGUAGE_NONE][0]['tid'], $compatible)) {
      continue;
    }
    $label = $release->versioncontrol_release['label'];
    $branches[$label['label_id']] = array(
      'name' => $label['name'],
      'label_id' => $label['label_id'],
      'release_nid' => $release->nid,
    );
  }

  // Get the vcs label_ids from the branches.
  $label_ids = array_keys($branches);

  // Use the label_ids to retrieve an array of test results.
  $test_results = array();
  $test_ids = pift_test_get_test_ids($label_ids, PIFT_TYPE_BRANCH);
  if (!empty($test_ids)) {
    $test_results = pift_test_get_summary_multiple($test_ids, PIFT_TYPE_BRANCH, TRUE);
    pift_rekey_by_value($test_results, 'nid');
    // Determine list of branches with releases but without tests.
    $no_results = array_diff($label_ids, array_keys($test_ids));
    $versions = array();
    foreach($no_results as $label_id) {
      if (!empty($branches[$label_id]['release_nid'])) {
        $versions[$label_id] = $branches[$label_id];
      }
    }
  }
  else {
    foreach ($branches as $label_id => $branch) {
      if (!empty($branches[$label_id]['release_nid'])) {
        $versions[$label_id] = $branches[$label_id];
      }
    }
  }

  $content['results_table_header'] = array(
    '#markup' => '<h3>' . t('Release Test Results') . '</h3>',
  );

  // Assemble the test results table render array.
  if (!empty($test_results)) {
    $content['results_table'] = array(
      '#theme' => 'table',
      '#header' => array(t('Release'), t('Status'), t('Result'), t('Last Tested'), t('Operations')),
      '#rows' => pift_pages_project_testing_rows($test_results, $branches),
      '#attached' => array(
        'css' => array(
          drupal_get_path('module', 'pift') . '/pift.css',
        ),
      ),
    );
  }
  else {
    $content['results_table'] = array(
      '#markup' => t("No test results available.  Please use the 'Test Additional Releases' form below to trigger a new test."),
    );
  }

  // Generate the 'Test Additional Branches' form.
  if (!empty($versions) && pift_pages_project_testing_settings_access($node)) {
    $content['testing_form'] = drupal_get_form('pift_pages_project_testing_form', $node, $versions);
  }

  return $content;
}

/**
 * Generate the results table #rows array for a given project node.
 *
 * @param array $test_ids
 *   An array of label_id => test_id pairings for the node's branches.
 * @param array $test_results
 *   A nested array of test result arrays.
 * @param string $branches
 *   An array of branch information, keyed by label id, as provided by
 *   pift_get_compatible_branches().
 *
 *  @return array $rows
 *    A nested array of table rows, suitable for feeding directly into the
 *    '#rows' property of theme_table().
 */
function pift_pages_project_testing_rows($test_results, $branches) {
  foreach ($branches as $label_id => $branch) {
    if (!empty($test_results[$branch['release_nid']])) {
      $rows[] = pift_pages_project_testing_row($test_results[$branch['release_nid']], $branch['name']);
    }
  }

  // Sort rows by branch name
  uasort($rows, 'pift_pages_sort_by_nested_value');

  return $rows;
}

/**
 * Sort callback for usort() with nested value.
 */
function pift_pages_sort_by_nested_value($a, $b) {
  return $a['data'][0] < $b['data'][0];
}

/**
 * Generates a single row of test results for the results table #rows array.
 *
 * @param array $test
 *   The test array for a single test.
 * @param string $version
 *   The human-readable version string for this particular branch.
 *
 *  @return $output
 *    An assembled table row element.
 */
function pift_pages_project_testing_row($test, $version) {
  if (!empty($test)) {
    $options = array('1' => 'Queued', '2' => 'Sent', '3' => 'Failed', '4' => 'Passed');
    $rowclass = array('1' => 'pift-bg-queue', '2' => 'pift-bg-retest', '3' => 'pift-bg-fail', '4' => 'pift-bg-pass');
    $operations = array();

    // A first-time queued test will not have a test_id, so hide links
    if ($test['test_id'] > 0) {
      $operations = array(
        pift_generate_view_link($test['test_id']),
      );
      if (user_access('pift re-test files') && $test['status'] > PIFT_STATUS_SENT) {
        $operations[] = pift_generate_retest_link($test['test_id']);
      }
      if (user_access('pift re-test files') && $test['status'] == PIFT_STATUS_SENT) {
        $operations[] = pift_generate_delete_link($test['test_id']);
      }
    }

    // Prepend 'queued' message to previous results if new test queued/sent
    if ($test['status'] <= PIFT_STATUS_SENT) {
      if (!empty($test['message'])) {
        $message = t('Waiting for results.') . "<br />";
        $message .= t('Previous test result: ') . $test['message'];
        $test['message'] = $message;
      }
      else {
        $test['message'] = t('Waiting for results.');
      }
    }

    // Format 'Last Tested' value
    $test['last_tested'] = $test['last_tested'] > 0 ? format_date($test['last_tested'], 'small') : 'never';

    // Assemble the row
    $row = array(
      'data' => array(check_plain($version), $options[$test['status']], $test['message'], $test['last_tested'], implode(" | ", $operations)),
      'class' => array($rowclass[$test['status']]),
    );
    return $row;
  }
  return array();
}

/**
 * Add additional description to file attachment form on issues.
 *
 * Invoked from within pift_form_project_issue_node_form_alter(), to modify the
 * file upload field description text.
 *
 * @see pift_form_project_issue_node_form_alter()
 */
function pift_pages_upload_description_add(&$form, $form_state, $form_id) {
  // Add description on field_issue_files if it is set.
  if (PIFT_DESCRIPTION) {
    $nid = $form['#node']->field_project[LANGUAGE_NONE][0]['target_id'];
    if (isset($form['field_issue_files'][LANGUAGE_NONE]['#file_upload_description']) && $nid && pift_project_enabled($nid)) {
      $description = filter_xss_admin(PIFT_DESCRIPTION);
      $form['field_issue_files'][LANGUAGE_NONE]['#file_upload_description'] .= " " . $description;
    }
  }
}

/**
 * Form constructor for the project testing preferences form on a project.
 *
 * @param object $node
 *   The fully loaded project node object.
 *
 * @see pift_pages_project_issue_settings_submit().
 */
function pift_pages_project_testing_settings($form, &$form_state, $node) {
  $form = array();
  $form['testing'] = array(
    '#type' => 'fieldset',
    '#title' => t('Testing'),
    '#collapsible' => TRUE,
    '#weight' => -9,
  );

  // Build an array of core capatible version strings (ie 6.x, 7.x, etc).
  $api_versions = pift_core_api_versions();
  $versions = array();
  foreach ($api_versions as $api_version) {
    $term = taxonomy_term_load($api_version);
    $versions[] = $term->name;
  }
  $form['testing']['pift_enable'] = array(
    '#type' => 'checkbox',
    '#title' => t('Enable automated testing'),
    '#description' => t('Enable automated testing of @versions compatible branches and patches.',
      array('@versions' => implode(', ', $versions))),
    '#default_value' => pift_project_enabled($node->nid),
    '#disabled' => !pift_pages_project_testing_settings_access($node),
  );
  $form['testing']['pid'] = array(
    '#type' => 'value',
    '#value' => $node->nid,
  );
  $form['testing']['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Save Settings'),
    '#access' => pift_pages_project_testing_settings_access($node),
  );
  return $form;
}

/**
 * Access callback for the project testing settings form on each project.
 */
function pift_pages_project_testing_settings_access($node) {
  global $user;
  return (user_access('access pift settings for all projects') ||
    (user_access('access pift settings for own projects')
     && !empty($node->project['maintainers'][$user->uid]['permissions']['write to vcs']))
   );
}

/**
 * Form submission handler for the pift_pages_project_testing_settings() form.
 *
 * Enables or disables testing for a given project.
 */
function pift_pages_project_testing_settings_submit($form, &$form_state) {
  $enabled = pift_project_enabled($form_state['values']['pid']);
  if ($form_state['values']['pift_enable'] && !$enabled) {
    pift_project_enable_testing($form_state['values']['pid']);
    drupal_set_message("Automated testing enabled.");
  }
  else if (!$form_state['values']['pift_enable'] && $enabled) {
    pift_project_disable_testing($form_state['values']['pid']);
    drupal_set_message("Automated testing disabled.");
  }
}

/**
 * Form constructor for the 'Test Additional Branches' form.
 */
function pift_pages_project_testing_form($form, &$form_state, $node, $versions) {
  $form = array();

  // Store the project nid and branches for use during submission processing.
  $form['pid'] = array(
    '#type' => 'value',
    '#value' => $node->nid,
  );
  $form['branches'] = array(
    '#type' => 'value',
    '#value' => $versions
  );
  // Form container

  $message = 'Marking a release for testing will add it to the list of branch/patch requests sent to the automated testing infrastructure. ';
  $message .= 'Once the tests complete, the results will be passed back to drupal.org and reflected on this page.';
  $description = '<p>' . $message . '</p>';

  $form['additional'] = array(
    '#type' => 'fieldset',
    '#title' => t('Test Additional Releases'),
    '#collapsible' => TRUE,
    '#collapsed' => FALSE,
    '#description' => $description,
  );
  foreach ($versions as $label_id => $version) {
    $options[$label_id] = $version['name'];
  }
  $form['additional']['version'] = array(
    '#type' => 'select',
    '#title' => t('Select a release to queue for testing'),
    '#options' => $options,
    '#required' => TRUE,
  );
  $form['additional']['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Queue Test'),
  );
  return $form;
}

/**
 * Form validation handler for the 'pift_pages_project_testing_form' form.
 */
function pift_pages_project_testing_form_validate($form, &$form_state) {
  // Ensure required values have been passed.
  if (empty($form_state['values']['pid']) || empty($form_state['values']['version'])) {
    form_set_error('form', t('Error: Unable to determine project branch.'));
  }
}

/**
 * Form submission handler for the 'pift_pages_project_testing_form' form.
 */
function pift_pages_project_testing_form_submit($form, &$form_state) {
  $branches = $form_state['values']['branches'];
  $label_id = $form_state['values']['version'];
  $release_nid = $branches[$label_id]['release_nid'];
  // Queue the test.
  pift_test_add(PIFT_TYPE_BRANCH, $label_id, $release_nid);
  drupal_set_message(t("The %version project branch has been queued for testing.", array('%version' => $branches[$label_id]['name'])));
}

/**
 * Form confirmation page for the 're-test' link associated with a file.
 *
 * @param integer $test_id
 *   The unique test ID to be re-tested.
 */
function pift_pages_retest_confirm_form($form, &$form_state, $test_id) {
  $test = pift_test_get_summary($test_id);

  if (!empty($test)) {
    $cid = pift_test_get_comment_id($test);
    $form = array();
    // Key off type to determine if this is a file retest or branch retest
    if ($test['type'] == PIFT_TYPE_FILE) {
      $file = file_load($test['id']);
      // File test
      $form['test'] = array(
        '#type' => 'value',
        '#value' => array(
          'test_id' => $test_id,
          'fid' => $test['id'],
          'nid' => $test['nid'],
          'cid' => $cid,
          'status' => $test['status'],
          'title' => $file->filename,
          'type' => PIFT_TYPE_FILE,
        ),
      );
      $form_state['#redirect'] = 'node/' . $form['test']['#value']['nid'];
    }
    elseif ($test['type'] == PIFT_TYPE_BRANCH) {
      // Branch test
      if (!empty($test['nid'])) {
        $release = node_load($test['nid']);
        $title = $release->title;
      }
      else {
        $title = t("Test %id", array('%id' => $test['test_id']));
      }

      $form['test'] = array(
        '#type' => 'value',
        '#value' => array(
          'test_id' => $test_id,
          'status' => $test['status'],
          'title' => $title,
          'type' => PIFT_TYPE_BRANCH,
        ),
      );
      // Get the project id from the release nid
      $pid = pift_test_get_pid($test);
      $form_state['#redirect'] = 'node/' . $pid . '/qa';
    }
    return confirm_form(
      $form,
      t('Are you sure you want to request that %title be re-tested?', array('%title' => $form['test']['#value']['title'])),
      $form_state['#redirect'],
      t('Once requested the test will be re-queued and results will be updated once the re-test has finished.'),
      t('Re-test'),
      t('Cancel')
    );
  }
  else {
    drupal_set_message(t('Invalid test.'), 'error');
  }
}

/**
 * Form submission handler for the pift_pages_retest_confirmation_form form.
 *
 * Requeue test if it still meets our criteria.
 */
function pift_pages_retest_confirm_form_submit($form, &$form_state) {
  $test = $form_state['values']['test'];
  if ($test['status'] > PIFT_STATUS_SENT) {
    if ($test['type'] == PIFT_TYPE_FILE) {
      // Base changes to be made in followup comment.
      $node = node_load($test['nid']);
      $account = user_load(variable_get('project_issue_followup_user', drupal_anonymous_user()));
      $node->nodechanges_uid = $account->uid;
      $node->nodechanges_comment['comment_body'][LANGUAGE_NONE][0] = array(
        'value' => theme('pift_auto_followup', array('type' => 'retest', 'nid' => $test['nid'], 'cid' => $test['cid'], 'filename' => $test['title'])),
        'format' => filter_default_format($account),
      );
      // If node issue status is not already a valid status, and this is the
      // last file in the field_issue_files stack, then set the default retest
      // status, otherwise leave status alone.
      $lastfile = end($node->field_issue_files['und']);
      if (!in_array($node->field_issue_status[LANGUAGE_NONE][0]['value'], variable_get('pift_status', array())) && $lastfile['fid'] == $test['fid']) {
        $node->field_issue_status[LANGUAGE_NONE][0]['value'] = PIFT_FOLLOWUP_RETEST;
      }
      // Clear out any null file placeholder within the field_issue_files
      // field, in order to prevent them from causing PHP notices. See
      // http://drupal.org/node/2061743.
      pift_sanitize_file_field($node);
      // Save the node.
      node_save($node);
    }
    // Update test record to reflect change.
    $result = pift_test_requeue($test['test_id']);

    if (!empty($result)) {
      drupal_set_message(t('%title has been submitted for re-testing. Please be patient while you wait for results.',
                           array('%title' => $test['title'])));
    }
    else {
      drupal_set_message(t('Invalid test.'), 'error');
    }
  }
  else {
    drupal_set_message(t('%title is not currently eligible for re-testing.',
                         array('%title' => $test['title'])), 'error');
  }
  $form_state['redirect'] = $form_state['#redirect'];
}

/**
 * Form confirmation page for a test deletion request.
 *
 * @param integer $test_id
 *   The unique test ID to be deleted.
 */
function pift_pages_delete_test_confirm_form($form, &$form_state, $test_id) {
  $test = pift_test_get_summary($test_id);

  if (!empty($test)) {

    if (($test['type'] == PIFT_TYPE_BRANCH) && !empty($test['nid'])) {
      $release = node_load($test['nid']);
      $title = $release->title;
    }
    else {
      $title = t("Test %id", array('%id' => $test['test_id']));
    }

    $form = array();
    $form['test'] = array(
      '#type' => 'value',
      '#value' => array(
        'test_id' => $test_id,
        'title' => $title,
      ),
    );

    // Set form redirect
    if ($test['type'] == PIFT_TYPE_BRANCH) {
      $pid = pift_test_get_pid($test);
      $form_state['#redirect'] = 'node/' . $pid . '/qa';
    }
    else {
      $form_state['#redirect'] = 'node/' . $test['nid'];
    }
    return confirm_form(
      $form,
      t('Are you sure you want to delete @title?', array('@title' => $form['test']['#value']['title'])),
      $form_state['#redirect'],
      t('This will remove any existing results from the site, and reset the test status.') . '<br />' . t('This can help free up a branch test which is stuck in a "sent" state for long periods of time.'),
      t('Delete Test'),
      t('Cancel')
    );
  }
  else {
    drupal_set_message(t('Invalid test.'), 'error');
  }
}

/**
 * Form submission handler for the pift_pages_delete_confirmation_form form.
 *
 * Delete the test.
 */
function pift_pages_delete_test_confirm_form_submit($form, &$form_state) {
  $test_id = $form_state['values']['test']['test_id'];
  if (!empty($test_id)) {
    pift_test_delete($test_id);
  }
  $form_state['redirect'] = $form_state['#redirect'];
}

/**
 * Process the specified attachment information and return the valid status
 * message and and row CSS class.
 *
 * @param integer $status
 *   Test status, PIFT_STATUS_*.
 * @param boolean $has_message
 *   Indicates whether the attachment has an associated message.
 *
 * @return array
 *   The table cell value and CSS class.
 */
function pift_attachment_process($status, $has_message) {
  $class = '';
  if ($status == PIFT_STATUS_QUEUE && $has_message) {
    $cell = t('Queued for re-testing');
    $class = 'pift-retest';
  }
  else if ($status == PIFT_STATUS_QUEUE) {
    $cell = t('Queued for testing');
  }
  elseif ($status == PIFT_STATUS_SENT) {
    $cell = t('Test request sent');
    $class = 'pift-retest';
  }
  elseif ($status > PIFT_STATUS_SENT) {
    $cell = t('Idle');
    $class = $status == PIFT_STATUS_PASS ? 'pift-pass' : 'pift-fail';
  }
  return array($cell, $class);
}
